"""Module containing the trait model."""
from typing import Tuple

import jax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float

from .typing import Population


class TraitModel:
    """Breeding simulator class. It can perform the most common operation of a breeding program.

    :param marker_effects: linear regressor weights for the traits.
    :type marker_effects: Array of shape (m, t), where t is the number of traits.
    :param offset: linear regressor offsets for each trait.
    :type offset: 1-dimensional array of length t or a float number.
    :param device: the device on which compute the traits estimation.
        If not specified, the default device will be chosen.
    :type device: XLA Device
    """

    def __init__(
        self,
        marker_effects: Float[Array, "m traits"],
        offset: Float[Array, "#traits"] = 0,
        device=None,
    ) -> None:
        """Initialize a trait model."""
        self.device = device
        self.marker_effects = jax.device_put(marker_effects, device=self.device)
        self.offset = jax.device_put(offset, device=self.device)
        self.n_traits = marker_effects.shape[1]

        props = _effect_properties(self.marker_effects, offset)
        self.positive_mask, self.max, self.min, self.mean, self.var = props

    def __call__(self, population: Population["n"]) -> Float[Array, "n traits"]:
        """Estimate the traits for the given population."""
        return _call(population, self.marker_effects, self.offset)


@jax.jit
def _call(
    population: Population["n"],
    marker_effects: Float[Array, "m traits"],
    offset: Float[Array, "#traits"],
) -> Float[Array, "n traits"]:
    monoploidy = population.sum(axis=-1, dtype=jnp.int8)
    return jnp.dot(monoploidy, marker_effects) + offset


@jax.jit
def _effect_properties(
    marker_effects: Float[Array, "m traits"], offset: Float[Array, "#traits"]
) -> Tuple[Bool[Array, "m"], float, float, float, float]:
    positive_mask = marker_effects > 0

    max_gebv = 2 * jnp.sum(marker_effects, axis=0, where=positive_mask) + offset
    min_gebv = 2 * jnp.sum(marker_effects, axis=0, where=~positive_mask) + offset
    mean = jnp.sum(marker_effects, axis=0) + offset
    # using variance property for sum of independent variables
    var = jnp.sum(marker_effects**2, axis=0) / 2

    return positive_mask, max_gebv, min_gebv, mean, var



import jax
import numpy as np
import pytest

from chromax import functional
from chromax.index_functions import conventional_index
from chromax.trait_model import TraitModel


@pytest.mark.parametrize("idx", [0, 1])
def test_cross(idx):
    n_markers, ploidy = 1000, 4
    n_crosses = 50
    parents_shape = (n_crosses, 2, n_markers, ploidy)
    parents = np.random.choice([False, True], size=parents_shape)
    rec_vec = np.zeros(n_markers)
    rec_vec[0] = idx
    random_key = jax.random.PRNGKey(42)
    new_pop = functional.cross(parents, rec_vec, random_key)

    for i in range(ploidy):
        assert np.all(new_pop[..., i] == parents[:, i % 2, :, i - i % 2 + idx])


def test_double_haploid():
    n_chr, chr_len, ploidy = 10, 100, 2
    n_offspring = 10
    pop_shape = (50, n_chr * chr_len, ploidy)
    f1 = np.random.choice([False, True], size=pop_shape)
    rec_vec = np.full((n_chr * chr_len,), 1.5 / chr_len)
    random_key = jax.random.PRNGKey(42)
    dh = functional.double_haploid(f1, n_offspring, rec_vec, random_key)
    assert dh.shape == (len(f1), n_offspring, n_chr * chr_len, ploidy)

    for i in range(ploidy // 2):
        assert np.all(dh[..., i * 2] == dh[..., i * 2 + 1])


def test_select():
    n_markers, ploidy = 1000, 4
    k = 10
    pop_shape = (50, n_markers, ploidy)
    f1 = np.random.choice([False, True], size=pop_shape)
    marker_effects = np.random.randn(n_markers)
    gebv_model = TraitModel(marker_effects[:, None])
    f_index = conventional_index(gebv_model)
    f2 = functional.select(f1, k=k, f_index=f_index)
    assert f2.shape == (k, *f1.shape[1:])

    f1_gebv = gebv_model(f1)
    f2_gebv = gebv_model(f2)
    assert np.max(f2_gebv) == np.max(f1_gebv)
    assert np.mean(f2_gebv) > np.mean(f1_gebv)
    assert np.min(f2_gebv) > np.min(f1_gebv)



"""Functional module."""
from functools import partial
from typing import Callable

import jax
import jax.numpy as jnp
from jaxtyping import Array, Float

from .typing import N_MARKERS, Haploid, Individual, Parents, Population


@jax.jit
def cross(
    parents: Parents["n"],
    recombination_vec: Float[Array, N_MARKERS],
    random_key: jax.random.PRNGKeyArray,
) -> Population["n"]:
    """Main function that computes crosses from a list of parents.

    :param parents: parents to compute the cross. The shape of
        the parents is (n, 2, m, d), where n is the number of parents,
        m is the number of markers, and d is the ploidy.
    :type parents: ndarray
    :param recombination_vec: array of m probabilities.
        The i-th value represent the probability to recombine before the marker i.
    :type recombination_vec:
    :param random_key: JAX PRNGKey, for reproducibility purpose.
    :type random_key: jax.random.PRNGKeyArray

    :return: offspring population of shape (n, m, d).
    :rtype: ndarray

    :Example:
        >>> from chromax import functional
        >>> import numpy as np
        >>> import jax
        >>> n_chr, chr_len, ploidy = 10, 100, 2
        >>> n_crosses = 50
        >>> parents_shape = (n_crosses, 2, n_chr * chr_len, ploidy)
        >>> parents = np.random.choice([False, True], size=parents_shape)
        >>> rec_vec = np.full((n_chr, chr_len), 1.5 / chr_len)
        >>> rec_vec[:, 0] = 0.5  # equal probability on starting haploid
        >>> rec_vec = rec_vec.flatten()
        >>> random_key = jax.random.PRNGKey(42)
        >>> f2 = functional.cross(parents, rec_vec, random_key)
        >>> f2.shape
        (50, 1000, 2)
    """
    parents = parents.reshape(*parents.shape[:3], -1, 2)
    random_keys = jax.random.split(random_key, num=len(parents) * 2 * parents.shape[3])
    random_keys = random_keys.reshape(len(parents), 2, parents.shape[3], 2)
    offsprings = _cross(parents, recombination_vec, random_keys)
    return offsprings.reshape(*offsprings.shape[:-2], -1)


@jax.jit
@partial(jax.vmap, in_axes=(0, None, 0))  # parallelize across individuals
@partial(jax.vmap, in_axes=(0, None, 0), out_axes=2)  # parallelize parents
def _cross(
    parent: Individual,
    recombination_vec: Float[Array, N_MARKERS],
    random_key: jax.random.PRNGKeyArray,
) -> Haploid:
    return _meiosis(parent, recombination_vec, random_key)


def double_haploid(
    population: Population["n"],
    n_offspring: int,
    recombination_vec: Float[Array, N_MARKERS],
    random_key: jax.random.PRNGKeyArray,
) -> Population["n n_offspring"]:
    """Computes the double haploid of the input population.

    :param population: input population of shape (n, m, d).
    :type population: ndarray
    :param n_offspring: number of offspring per plant.
    :type n_offspring: int
    :param recombination_vec: array of m probabilities.
        The i-th value represent the probability to recombine before the marker i.
    :type recombination_vec: ndarray
    :param random_key: array of n PRNGKey, one for each individual.
    :type random_key: jax.random.PRNGKeyArray

    :return: output population of shape (n, n_offspring, m, d).
        This population will be homozygote.
    :rtype: ndarray

    :Example:
        >>> from chromax import functional
        >>> import numpy as np
        >>> import jax
        >>> n_chr, chr_len, ploidy = 10, 100, 2
        >>> pop_shape = (50, n_chr * chr_len, ploidy)
        >>> f1 = np.random.choice([False, True], size=pop_shape)
        >>> rec_vec = np.full((n_chr, chr_len), 1.5 / chr_len)
        >>> rec_vec[:, 0] = 0.5  # equal probability on starting haploid
        >>> rec_vec = rec_vec.flatten()
        >>> random_key = jax.random.PRNGKey(42)
        >>> dh = functional.double_haploid(f1, 10, rec_vec, random_key)
        >>> dh.shape
        (50, 10, 1000, 2)
    """
    population = population.reshape(*population.shape[:2], -1, 2)
    keys = jax.random.split(
        random_key, num=len(population) * n_offspring * population.shape[2]
    ).reshape(len(population), n_offspring, population.shape[2], 2)
    haploids = _double_haploid(population, recombination_vec, keys)
    dh_pop = jnp.broadcast_to(haploids[..., None], shape=(*haploids.shape, 2))
    return dh_pop.reshape(*dh_pop.shape[:-2], -1)


@jax.jit
@partial(jax.vmap, in_axes=(0, None, 0))  # parallelize across individuals
@partial(jax.vmap, in_axes=(None, None, 0))  # parallelize across offsprings
def _double_haploid(
    individual: Individual,
    recombination_vec: Float[Array, N_MARKERS],
    random_key: jax.random.PRNGKeyArray,
) -> Haploid:
    return _meiosis(individual, recombination_vec, random_key)


@jax.jit
@partial(jax.vmap, in_axes=(1, None, 0), out_axes=1)  # parallelize pair of chromosomes
def _meiosis(
    individual: Individual,
    recombination_vec: Float[Array, N_MARKERS],
    random_key: jax.random.PRNGKeyArray,
) -> Haploid:
    samples = jax.random.uniform(random_key, shape=recombination_vec.shape)
    rec_sites = samples < recombination_vec
    crossover_mask = jax.lax.associative_scan(jnp.logical_xor, rec_sites)

    crossover_mask = crossover_mask.astype(jnp.int8)
    haploid = jnp.take_along_axis(individual, crossover_mask[:, None], axis=-1)

    return haploid.squeeze()


def select(
    population: Population["n"],
    k: int,
    f_index: Callable[[Population["n"]], Float[Array, "n"]],
) -> Population["k"]:
    """Function to select individuals based on their score (index).

    :param population: input grouped population of shape (n, m, d)
    :type population: ndarray
    :param k: number of individual to select.
    :type k: int
    :param f_index: function that computes a score for each individual.
        The function accepts as input a population, i.e. and array of shape
        (n, m, 2) and returns an array of n float number.
    :type f_index: Callable

    :return: output population of (k, m, d)
    :rtype: ndarray

    :Example:
        >>> from chromax import functional
        >>> from chromax.trait_model import TraitModel
        >>> from chromax.index_functions import conventional_index
        >>> import numpy as np
        >>> n_chr, chr_len, ploidy = 10, 100, 2
        >>> pop_shape = (50, n_chr * chr_len, ploidy)
        >>> f1 = np.random.choice([False, True], size=pop_shape)
        >>> marker_effects = np.random.randn(n_chr * chr_len)
        >>> gebv_model = TraitModel(marker_effects[:, None])
        >>> f_index = conventional_index(gebv_model)
        >>> f2 = functional.select(f1, k=10, f_index=f_index)
        >>> f2.shape
        (10, 1000, 2)
    """
    indices = f_index(population)
    _, best_pop = jax.lax.top_k(indices, k)
    return population[best_pop, :, :]



"""Utility module with common index functions."""
from math import ceil
from typing import Callable

import numpy as np
from jaxtyping import Array, Float

from chromax.trait_model import TraitModel
from chromax.typing import Population


def phenotype_index(simulator, environments: np.ndarray) -> Callable:
    """Function to select based on phenotyping with some environments.

    :param simulator: chromax simulator instance to use for phenotyping simulation.
    :type simulator: chromax.Simulator
    :param environments: environments created using `create_environments` method from the simulator.

    :return: the phenotyping index function to use for selection.
    :rtype: Callable[[Population["n"]], Float[Array, "n"]]
    """

    def phenotype_index_f(population: Population["n"]) -> Float[Array, "n"]:
        phenotype = simulator.phenotype(
            population, environments=environments, raw_array=True
        )
        return phenotype[..., 0]

    return phenotype_index_f


def visual_selection(simulator, noise_factor: int = 1, seed: int = None) -> Callable:
    """Function to select based on visual selection.

    Practically, this is similar to phenotyping but with more noise.

    :param simulator: chromax simulator instance to use for phenotyping simulation.
    :type simulator: chromax.Simulator
    :param noise_factor: variance ratio between the phenotype and artificial noise added
        to simulate visual selection inaccuracy.
    :type noise_factor: int
    :param seed: random seed for reproducibility.
    :type seed: int

    :return: the visual selection index function.
    :rtype: Callable[[Population["n"]], Float[Array, "n"]]
    """
    generator = np.random.default_rng(seed)

    def visual_selection_f(population: Population["n"]) -> Float[Array, "n"]:
        phenotype = simulator.phenotype(population, raw_array=True)[..., 0]
        noise_var = (simulator.GxE_model.var + simulator.GxE_model.var) * noise_factor
        noise = generator.normal(scale=np.sqrt(noise_var), size=phenotype.shape)
        return phenotype + noise

    return visual_selection_f


def conventional_index(GEBV_model: TraitModel):
    """Function to select based on Genomic Estimated Breeding Value (GEBV).

    :param GEBV_model: GEBV model to estimate the genomic breeding value.
        It must return a single value for an individual, i.e. estimate a single trait.
    :type GEBV_model: chromax.TraitModel

    :return: the conventional genomic selection index function.
    :rtype: Callable[[Population["n"]], Float[Array, "n"]]
    """

    def conventional_index_f(pop: Population["n"]) -> Float[Array, "n"]:
        gebv = GEBV_model(pop)
        assert gebv.shape[-1] == 1
        return gebv[..., 0]

    return conventional_index_f


def _poor_conventional_indices(GEBV_model, pop, F):
    n_poors = int(len(pop) * F)
    GEBVs = conventional_index(GEBV_model)(pop)
    sorted_indices = np.argsort(GEBVs)
    poor_indices = sorted_indices[:n_poors]
    return poor_indices


def optimal_haploid_value(GEBV_model, F=0, B=None, chr_lens=None):
    """Method implementing Optimal Haploid Value (OHV) index function."""

    def optimal_haploid_value_f(pop):
        OHP = optimal_haploid_pop(GEBV_model, pop, B, chr_lens)
        OHV = 2 * GEBV_model(OHP[..., None]).squeeze()
        OHV = np.array(OHV)

        if F > 0:
            remove_indices = _poor_conventional_indices(GEBV_model, pop, F)
            OHV[remove_indices] = -float("inf")

        return OHV

    return optimal_haploid_value_f


def optimal_haploid_pop(GEBV_model, population, B=None, chr_lens=None):
    """Function returning the optimal haploid of a population."""
    if B is None:
        return _optimal_haploid_pop(GEBV_model, population)
    else:
        assert chr_lens is not None, "chr_lens needed with B"
        return _optimal_haploid_pop_B(GEBV_model, population, B, chr_lens)


def _optimal_haploid_pop_B(GEBV_model, population, B, chr_lens):
    OHP = np.empty((population.shape[0], population.shape[1]), dtype="bool")

    start_idx = 0
    for chr_length in chr_lens:
        block_length = ceil(chr_length / B)
        end_chr = start_idx + chr_length
        for _ in range(B):
            end_block = min(start_idx + block_length, end_chr)
            pop_slice = population[:, start_idx:end_block]
            effect_slice = GEBV_model.marker_effects[start_idx:end_block]
            block_gebv = np.einsum("nml,me->nl", pop_slice, effect_slice)
            best_blocks = np.argmax(block_gebv, axis=-1)
            OHP[:, start_idx:end_block] = np.take_along_axis(
                pop_slice, best_blocks[:, None, None], axis=2
            ).squeeze(axis=-1)
            start_idx = end_block

        assert start_idx == end_chr

    return OHP


def _optimal_haploid_pop(GEBV_model, population):
    optimal_haploid_pop = np.empty(
        (population.shape[0], population.shape[1]), dtype="bool"
    )

    positive_mask = GEBV_model.positive_mask.squeeze()

    optimal_haploid_pop[:, positive_mask] = np.any(
        population[:, positive_mask], axis=-1
    )
    optimal_haploid_pop[:, ~positive_mask] = np.all(
        population[:, ~positive_mask], axis=-1
    )

    return optimal_haploid_pop


def optimal_population_value(GEBV_model, n, F=0, B=None, chr_lens=None):
    """Method implementing Optimal Population Value (OPV) index function."""

    def optimal_population_value_f(population):
        indices = np.arange(len(population))
        remove_indices = _poor_conventional_indices(GEBV_model, population, F)
        indices = np.delete(indices, remove_indices)

        output = np.zeros(len(population), dtype="bool")
        positive_mask = GEBV_model.marker_effects[:, 0] > 0
        current_set = ~positive_mask
        G = optimal_haploid_pop(GEBV_model, population[indices], B, chr_lens)

        for _ in range(n):
            dummy_pop = np.broadcast_to(current_set[None, :], G.shape)
            dummy_pop = np.stack((G, dummy_pop), axis=2)
            G = optimal_haploid_pop(GEBV_model, dummy_pop, B, chr_lens)
            best_ind = np.argmax(GEBV_model(G[:, :, None]))
            output[indices[best_ind]] = True
            current_set = G[best_ind]

            indices = np.delete(indices, best_ind)
            G = optimal_haploid_pop(GEBV_model, population[indices], B, chr_lens)

        assert np.count_nonzero(output) == n
        return output  # not OPV but a mask

    return optimal_population_value_f



"""Module containing simulator class."""
import logging
import random
from pathlib import Path
from typing import Callable, List, Optional, Union

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax._src.lib import xla_client as xc
from jaxtyping import Array, Float, Int

from . import functional
from .index_functions import conventional_index
from .trait_model import TraitModel
from .typing import Parents, Population


class Simulator:
    """Breeding simulator class. It can perform the most common operation of a breeding program.

    :param genetic_map: the path, or dataframe, containing the genetic map.
        It needs to have all the columns specified in trait_names, `CHR.PHYS`
        (with the name of the marker chromosome), and one between `cM` or `RecombRate`.
    :type genetic_map: Path or DataFrame
    :param trait_names: column names in the genetic_map.
        The values of the columns are the marker effects on the trait for each marker.
        The default value is `Yield`.
    :type trait_names: List of strings
    :param chr_column: name of the column containing the chromosome identifier.
        The default value is `CHR.PHYS`.
    :type chr_column: str
    :param position_column: name of the column containing the position in cM of the marker.
        The default value is `cM`.
    :type position_column: str
    :param recombination_column: name of the column containing the probability that a
        recombination happens before the current marker and after the previous one.
        The default value is `RecombRate`.
    :type recombination_column: str
    :param h2: narrow-sense heritability value for each trait.
        The default value is 0.5 for each trait.
    :type h2: array of float
    :param seed: the random seed for reproducibility.
    :type seed: int
    :param device: the device for computing simulations. It will be automatically selected if not
        specified; by default to the first available GPU or TPU, or the CPU if neither is present.
    :type device: XLA Device
    :param backend: the backend of the device.
        Common choices are `gpu`, `cpu` or `tpu`.
    :type backend: str or XLA client

    :Example:
        >>> from chromax import Simulator, sample_data
        >>> simulator = Simulator(genetic_map=sample_data.genetic_map)
        >>> f1 = simulator.load_population(sample_data.genome)
        >>> f2, _ = simulator.random_crosses(f1, n_crosses=10, n_offspring=20)
        >>> f2.shape
        (10, 20, 9839, 2)
    """

    def __init__(
        self,
        genetic_map: Union[Path, pd.DataFrame],
        trait_names: Optional[List[str]] = None,
        chr_column: str = "CHR.PHYS",
        position_column: str = "cM",
        recombination_column: str = "RecombRate",
        h2: Optional[np.ndarray] = None,
        seed: Optional[int] = None,
        device: xc.Device = None,
        backend: Union[str, xc._xla.Client] = None,
    ):
        """Initialization method. See class docstring for information about parameters."""
        self.random_key = None
        if seed is None:
            seed = random.randint(0, 2**32)
        self.set_seed(seed)

        if device is None:  # use the right backend
            device = jax.local_devices(backend=backend)[0]
        elif isinstance(device, int):
            local_devices = jax.local_devices(backend=backend)
            matched_devices = filter(lambda d: d.id == device, local_devices)
            matched_devices = list(matched_devices)
            assert len(matched_devices) <= 1
            if len(matched_devices) == 0:
                print(jax.devices(), flush=True)
                logging.warning("No device with id: %i. Using the default one.", device)
                device = jax.local_devices(backend=backend)[0]
            else:
                device = matched_devices[0]
        self.device = device

        if not isinstance(genetic_map, pd.DataFrame):
            genetic_map = pd.read_table(genetic_map, sep="\t")
        if trait_names is None:
            other_col = {chr_column, position_column, recombination_column}
            trait_names = genetic_map.columns.drop(other_col, errors="ignore")
        self.trait_names = trait_names

        self.n_markers = len(genetic_map)
        chr_map = genetic_map[chr_column]
        self.chr_lens = chr_map.groupby(chr_map).count().values

        mrk_effects = genetic_map[self.trait_names]
        self.GEBV_model = TraitModel(
            marker_effects=mrk_effects.to_numpy(dtype=np.float32), device=self.device
        )

        if h2 is None:
            h2 = np.full((len(self.trait_names),), 0.5)
        h2 = np.asarray(h2)
        self.random_key, split_key = jax.random.split(self.random_key)
        env_effects = jax.random.normal(
            split_key, shape=(self.n_markers, len(self.trait_names))
        )
        target_vars = (1 - h2) / h2 * self.GEBV_model.var
        env_effects *= np.sqrt(2 * target_vars / self.n_markers)
        self.GxE_model = TraitModel(
            marker_effects=env_effects, offset=1, device=self.device
        )

        if recombination_column in genetic_map.columns:
            recombination_vec = genetic_map[recombination_column].to_numpy(
                dtype=np.float32
            )
            # change semantic to "recombine now" instead of "recombine after"
            recombination_vec[1:] = recombination_vec[:-1]
            self.cM = np.zeros(self.n_markers, dtype=np.float32)
            start_idx = 0
            for chr_len in self.chr_lens:
                end_idx = start_idx + chr_len
                self.cM[start_idx + 1 : end_idx] = (
                    recombination_vec[start_idx + 1 : end_idx].cumsum() * 100
                )
                start_idx = end_idx

        elif position_column in genetic_map.columns:
            self.cM = genetic_map[position_column].to_numpy(dtype=np.float32)
            recombination_vec = np.zeros(self.n_markers, dtype=np.float32)
            start_idx = 0
            for chr_len in self.chr_lens:
                end_idx = start_idx + chr_len
                recombination_vec[start_idx + 1 : end_idx] = (
                    self.cM[start_idx + 1 : end_idx] - self.cM[start_idx : end_idx - 1]
                )
                recombination_vec /= 100
        else:
            raise ValueError(
                f"One between {recombination_column} and {position_column} must be specified"
            )

        first_mrk_map = np.zeros(len(chr_map), dtype="bool")
        first_mrk_map[1:] = chr_map.iloc[1:].values != chr_map.iloc[:-1].values
        first_mrk_map[0] = True
        recombination_vec[first_mrk_map] = 0.5  # first equally likely
        self.recombination_vec = jax.device_put(recombination_vec, device=self.device)

    def set_seed(self, seed: int):
        """Set random seed for reproducibility.

        :param seed: random seed.
        :type seed: int
        """
        self.random_key = jax.random.PRNGKey(seed)

    def load_population(self, file_name: Union[Path, str]) -> Population["n"]:
        """Load a population from file.

        :param file_name: path of the file with the population genome.
        :type file_name: path

        :return: loaded population of shape (n, m, d), where
            n is the number of individual, m is the total number of marker,
            and d is the diploidy of the population.
        :rtype: ndarray

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map)
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> f1.shape
            (371, 9839, 2)
        """
        population = np.load(file_name)
        return jax.device_put(population, device=self.device)

    def save_population(self, population: Population["n"], file_name: Union[Path, str]):
        """Save a population to file.

        :param population: population to save.
        :type population: ndarray
        :file_name: file path to save the population.
        :type file_name: path

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map)
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> f2, _ = simulator.random_crosses(f1, n_crosses=10, n_offspring=20)
            >>> simulator.save_population(f2, "pop_file")
        """
        np.save(file_name, population, allow_pickle=False)

    def cross(self, parents: Parents["n"]) -> Population["n"]:
        """Main function that computes crosses from a list of parents.

        :param parents: parents to compute the cross. The shape of
            the parents is (n, 2, m, d), where n is the number of parents,
            m is the number of markers, and d is the ploidy.
        :type parents: ndarray


        :return: offspring population of shape (n, m, d).
        :rtype: ndarray

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> import numpy as np
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map)
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> parents_indices = np.array([
                [1, 5],
                [4, 7],
                [5, 6]
            ])
            >>> parents = f1[parents_indices]
            >>> f2 = simulator.cross(parents)
            >>> f2.shape
            (3, 9839, 2)
        """
        self.random_key, split_key = jax.random.split(self.random_key)
        return functional.cross(parents, self.recombination_vec, split_key)

    @property
    def differentiable_cross_func(self) -> Callable:
        """Experimental features that return a differentiable version of the cross function.

        The differentiable crossing function takes as input:
         - population (array): starting population from which performing the crosses.
            The shape of the population is (n, m, d).
         - cross_weights (array): Array of shape (l, n, d). It is used to compute
            l crosses, starting from a weighted average of the n possible parents.
            When the n-axis has all zeros except of a single element equals to one,
            this function is equivalent to the cross function.
         - random_key (JAX random key): random key used for recombination sampling.

        And returns a population of shape (l, m, d).

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> import numpy as np
            >>> import jax
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map)
            >>> diff_cross = simulator.differentiable_cross_func
            >>> def mean_gebv(pop, weights, random_key):
                    new_pop = diff_cross(pop, weights, random_key)
                    return simulator.GEBV(new_pop, raw_array=True).mean()
            >>> grad_f = jax.grad(mean_gebv, argnums=1)
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> weights = np.random.uniform(size=(10, len(f1), 2))
            >>> weights /= weights.sum(axis=1, keepdims=True)
            >>> random_key = jax.random.PRNGKey(42)
            >>> grad_value = grad_f(f1, weights, random_key)
            >>> grad_value.shape
            (10, 371, 2)
        """
        cross_haplo = jax.vmap(functional._meiosis, in_axes=(None, None, 0), out_axes=1)
        cross_individual = jax.vmap(cross_haplo, in_axes=(0, None, 0))
        cross_pop = jax.vmap(cross_individual, in_axes=(None, None, 0))

        @jax.jit
        def diff_cross_f(
            population: Population["n"],
            cross_weights: Float[Array, "m n 2"],
            random_key: jax.random.PRNGKeyArray,
        ) -> Population["m"]:
            population = population.reshape(*population.shape[:-1], -1, 2)
            keys_shape = len(cross_weights), len(population), 2, population.shape[-2]
            keys = jax.random.split(random_key, num=np.prod(keys_shape))
            keys = keys.reshape(*keys_shape, 2)
            outer_res = cross_pop(population, self.recombination_vec, keys)
            outer_res = outer_res.reshape(*outer_res.shape[:-2], -1)
            return (cross_weights[:, :, None, :] * outer_res).sum(axis=1)

        return diff_cross_f

    def double_haploid(
        self, population: Population["n"], n_offspring: int = 1
    ) -> Population["n n_offspring"]:
        """Computes the double haploid of the input population.

        :param population: input population of shape (n, m, 2).
        :type population: ndarray
        :param n_offspring: number of offspring per plant.
            The default value is 1.
        :type n_offspring: int

        :return: output population of shape (n, n_offspring, m, 2).
            This population will be homozygote.
        :rtype: ndarray

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map)
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> dh = simulator.double_haploid(f1, n_offspring=10)
            >>> dh.shape
            (371, 10, 9839, 2)
        """
        self.random_key, split_key = jax.random.split(self.random_key)
        dh = functional.double_haploid(
            population, n_offspring, self.recombination_vec, split_key
        )

        if n_offspring == 1:
            dh = dh.squeeze(1)
        return dh

    def diallel(
        self, population: Population["n"], n_offspring: int = 1
    ) -> Population["n*(n-1)/2 n_offspring"]:
        """Diallel crossing function (crossing between every possible couple) except self-crossing.

        :param population: input population of shape (n, m, d).
        :type population: ndarray
        :param n_offspring: number of offspring per cross.
            The default value is 1.
        :type n_offspring: int

        :return: output population of shape (l, n_offspring, m, d),
            where l is the number of possible pair, i.e `n * (n-1) / 2`.
        :rtype: ndarray

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map)
            >>> f1 = simulator.load_population(sample_data.genome)[:10]
            >>> f2 = simulator.diallel(f1, n_offspring=10)
            >>> f2.shape
            (45, 10, 9839, 2)
        """
        if n_offspring < 1:
            raise ValueError("n_offspring must be higher or equal to 1")

        all_indices = jnp.arange(len(population))
        diallel_indices = self._diallel_indices(all_indices)
        cross_indices = jnp.repeat(diallel_indices, n_offspring, axis=0)
        out = self.cross(population[cross_indices])
        out = out.reshape(len(diallel_indices), n_offspring, *out.shape[1:])
        if n_offspring == 1:
            out = out.squeeze(1)
        return out

    def _diallel_indices(self, indices: Int[Array, "n"]) -> Int[Array, "n*(n-1)/2"]:
        triu_indices = jnp.triu_indices(len(indices), k=1)
        mesh1 = indices[triu_indices[0]]
        mesh2 = indices[triu_indices[1]]
        return jnp.stack([mesh1, mesh2], axis=1)

    def random_crosses(
        self, population: Population["n"], n_crosses: int, n_offspring: int = 1
    ) -> Population["n_crosses n_offspring"]:
        """Computes random crosses on a population.

        :param population: input population of shape (n, m, d).
        :type population: ndarray
        :param n_crosses: number of random crosses to perform.
        :type n_crosses: int
        :param n_offspring: number of offspring per cross.
            The default value is 1.
        :type n_offspring: int

        :return: output population of shape (n_crosses, n_offspring, m, d)
            and parent indices of shape (n_crosses, 2) of performed crosses.
        :rtype: tuple of two ndarrays

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map)
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> f2, parent_ids = simulator.random_crosses(f1, 100, n_offspring=10)
            >>> f2.shape
            (100, 10, 9839, 2)
            >>> parent_ids.shape
            (100, 2)
        """
        all_indices = np.arange(len(population))
        diallel_indices = self._diallel_indices(all_indices)
        if n_crosses > len(diallel_indices):
            raise ValueError("n_crosses can be at most the diallel length")

        self.random_key, split_key = jax.random.split(self.random_key)
        random_select_idx = jax.random.choice(
            split_key, len(diallel_indices), shape=(n_crosses,), replace=False
        )
        parent_indices = diallel_indices[random_select_idx]

        cross_indices = np.repeat(parent_indices, n_offspring, axis=0)
        out = self.cross(population[cross_indices])
        out = out.reshape(n_crosses, n_offspring, *out.shape[1:])
        if n_offspring == 1:
            out = out.squeeze(1)
        return out, parent_indices

    def select(
        self,
        population: Population["_g n"],
        k: int,
        f_index: Optional[Callable[[Population["n"]], Float[Array, "n"]]] = None,
    ) -> Population["_g k"]:
        """Function to select individuals based on their score (index).

        :param population: input population of shape (n, m, d),
            or shape (g, n, m, d), to select k individual from each group population group g.
        :type population: ndarray
        :param k: number of individual to select.
        :type k: int
        :param f_index: function that computes a score from each individual.
            The function accepts as input the population, i.e. and array of shape
            (n, m, d) and returns a n float numbers. The default f_index is the conventional index,
            i.e. the sum of the marker effects masked with the SNPs from the genetic_map.
        :type f_index: Callable

        :return: output population of shape (k, m, d) or (g, k, m, d),
            depending on the input population.
        :rtype: ndarray

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map, trait_names=["Yield"])
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> len(f1), simulator.GEBV(f1).mean().values
            (371, [8.223844])
            >>> f2 = simulator.select(f1, k=20)
            >>> len(f2), simulator.GEBV(f2).mean().values
            (20, [14.595136])
        """
        if f_index is None:
            f_index = conventional_index(self.GEBV_model)

        if len(population.shape) == 3:
            select_f = functional.select
        elif len(population.shape) == 4:
            select_f = jax.vmap(functional.select, in_axes=(0, None, None))
        else:
            raise ValueError(
                f"Unexpected shape {population.shape} for input population"
            )

        return select_f(population, k, f_index)

    def GEBV(
        self, population: Population["n"], *, raw_array: bool = False
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Computes the Genomic Estimated Breeding Values using the data from the genetic_map.

        :param population: input population of shape (n, m, d).
        :type population: ndarray
        :param raw_array: whether to return a raw array or a DataFrame.
            Default value is False.
        :type raw_array: bool

        :return: a DataFrame (or array) with n rows and a column for each trait.
            It contains the GEBV of each trait for each individual.
        :rtype: DataFrame or ndarray

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map)
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> simulator.GEBV(f1).mean()
            Heading Date              0.196119
            Protein Content          -0.228718
            Plant Height             -5.888406
            Thousand Kernel Weight   -1.029418
            Yield                     8.223843
            Fusarium Head Blight      5.318052
            Spike Emergence Period   -0.933169
            dtype: float32
        """
        GEBV = self.GEBV_model(population)
        if not raw_array:
            GEBV = pd.DataFrame(GEBV, columns=self.trait_names)
        return GEBV

    def create_environments(
        self, num_environments: int
    ) -> Float[Array, "num_environments"]:
        """Create environments to phenotype the population.

        In practice, it generates random numbers from a normal distribution.

        :param num_environments: number of environments to create.
        :type num_environments: int

        :return: array of floating point numbers.
            This output can be used for the function `phenotype`.
        :rtype: ndarray
        """
        self.random_key, split_key = jax.random.split(self.random_key)
        return jax.random.normal(
            key=split_key,
            shape=(num_environments,),
        )

    def phenotype(
        self,
        population: Population["n"],
        *,
        num_environments: Optional[int] = None,
        environments: Optional[np.ndarray] = None,
        raw_array: bool = False,
    ) -> Union[pd.DataFrame, np.ndarray]:
        """Simulates the phenotype of a population.

        This uses the Genotype-by-Environment model described in `AlphaSimR
        <https://cran.r-project.org/web/packages/AlphaSimR/vignettes/traits.pdf>`_.

        :param population: input population of shape (n, m, d)
        :type population: ndarray
        :param num_environments: number of environments to test the population.
            Default value is 1.
        :type num_environments: int
        :param environments: environments to test the population. Each environment
            must be represented by a floating number in the range (-1, 1).
            When drawing new environments use normal distribution to maintain
            heretability semantics.
        :type environments: ndarray
        :param raw_array: whether to return a raw array or a DataFrame.
            Default value is False.
        :type raw_array: bool

        :return: a DataFrame (or array) with n rows and a column for each trait.
            It contains the simulated phenotype for each individual.
        :rtype: DataFrame or ndarray

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map, seed=42)
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> envs = simulator.create_environments(4)
            >>> simulator.phenotype(f1, environments=envs).mean()
            Heading Date              0.105397
            Protein Content          -0.172026
            Plant Height             -5.813669
            Thousand Kernel Weight   -1.372738
            Yield                     8.306302
            Fusarium Head Blight      4.286477
            Spike Emergence Period   -0.575061
            dtype: float32
        """
        if num_environments is not None and environments is not None:
            raise ValueError(
                "You cannot specify both num_environments and environments"
            )
        if environments is None:
            num_environments = num_environments if num_environments is not None else 1
            environments = self.create_environments(num_environments)

        w = jnp.mean(environments)
        GEBV = self.GEBV_model(population)
        GxE = self.GxE_model(population)
        phenotype = GEBV + w * GxE
        if not raw_array:
            phenotype = pd.DataFrame(phenotype, columns=self.trait_names)
        return phenotype

    def corrcoef(self, population: Population["n"]) -> Float[Array, "n"]:
        """Computes the correlation coefficient of the population against its centroid.

        It can be used as an indicator of variance in the population.

        :param population: input population of shape (n, m, d)
        :type population: ndarray

        :return: vector of length n, containing the correlation coefficient
            of each individual against the average of the population.
        :rtype: ndarray

        :Example:
            >>> from chromax import Simulator, sample_data
            >>> simulator = Simulator(genetic_map=sample_data.genetic_map, seed=42)
            >>> f1 = simulator.load_population(sample_data.genome)
            >>> corrcoef = simulator.corrcoef(f1)
            >>> corrcoef.shape
            (371,)
        """
        monoploid_enc = population.reshape(population.shape[0], -1)
        mean_pop = jnp.mean(monoploid_enc, axis=0)
        pop_with_centroid = jnp.vstack([mean_pop, monoploid_enc])
        corrcoef = jnp.corrcoef(pop_with_centroid)
        return corrcoef[0, 1:]
